# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

#
# Test for IP behavior
#

package ActiveMap::IP;

use strict;
use Net::RawIP;
use ActiveMap::Utils;


#
# Globals
#



# Default filter is to look for ICMP echo replies; any filters provided
# for a given tests are ORed with the default filter.
my %ip_tests =
  (
#     bad_checksum => {
#  		    control_fn => \&bad_checksum,
#  		    repeat_count => 3,  # how many times to do the test
#  		    expected_pkt_len => 100, # max response pkt size expected
#  		    # in ms, how long before the test ends (regardless of 'done' status)
#  		    test_timeout => 3,
#  		    filter => undef,
#  		   },

   overlapping_frag => {
			control_fn => \&overlapping_frag,
			repeat_count => 3,
			expected_pkt_len => 300,
			test_timeout => 5,
			filter => undef
		       }
  );


# hostname => { name of test => array of test results (one per iteration) }
# (each result is a hashref)
my %results;

# target.testname =>
#  {
#    target => endhost being tested
#    testname => name of the test
#    time => time test was started (UNIX time),
#    iters_left => num. remaining iterations,
#    result_ref => \%curr_result,
#    icmp_id => ID field
#    done_flag => true if test should be ended (result already obtained)
#  }
my %active_tests;

my %id_map; # { icmp_code => active_test }

my $g_next_icmp_id = int (rand (2**16));

sub get_next_icmp_id { 
  $g_next_icmp_id++; 
  $g_next_icmp_id -= 2**16 if $g_next_icmp_id >= 2**16; 
  return $g_next_icmp_id;
}

#
# Individual Tests
#
# Each test has two functions, one for sending query packets and one for
# interpreting the results.
# The receiving function can return one of a number of status codes:
# undef for no change, 'done' for test complete, 'restart' for
# restarting the test if results were inconclusive--care should be
# taken to avoid infinite loops..
#



#
# $args = { state => $active_tests{thistest},
#           timeout => 1 if this test timed out -- results should be written
#           hdr => $hdr, pkt => Net::RawIP obj }
#
#
sub bad_checksum
{
    my $args = shift;

    if ($args->{timeout})
      {
	$args->{state}{result_ref}{allows_bad_checksum} = 0;
	$args->{done_flag} = 1;
	return;
      }

    if (! $args->{pkt})
      {
	# initial call
	my $request = new Net::RawIP
	  ({
	    ip   => { saddr => ActiveMap::Utils::this_ip(), daddr => $args->{state}{target},
		      check => 2 }, # *** small chance it will match...
	    icmp => { type => 8, code => 0, id => $args->{state}{icmp_id}, check => 0 }
	   });
	$request->send();
	return;
      }

    $args->{state}{result_ref}{allows_bad_checksum} = 1;
    $args->{done_flag} = 1;
}


#
# Send the following 6 fragments (the sixth one prevents reassembly until the end)
# 012345678901 IP Offset (8 bytes/character)
# 111 22333
#  4444 555666
#
# If we have a 'first' policy:     111422333666
# If we have a 'last' policy:      144442555666
# If we have the Linux policy:     111442555666
# If we have a 'BSD'-style policy: 111442333666
# For 'BSD' with right-trim:       144422555666
# Anything else is new and should be investigated. 
#
sub overlapping_frag
  {
    my $args = shift;

    my %policies = ( 
		      'first'     => '111422333666',
		       'last'      => '144442555666',
		     'BSD'       => '111442333666',
		     'linux'     => '111442555666',
  		     'unknown3'  => '144442333666',
		     'unknown1'  => '111422555666',
  		     'unknown2'  => '144422333666',
		     'BSD-right' => '144422555666' 
		   );

    if ($args->{timeout})
      {
	if (! keys %{$args->{state}{result_ref}})
	  {
	    $args->{state}{result_ref}{overlap_policy} = 'No result: timeout or unknown policy';
	    $args->{done_flag} = 1;
	    return;
	  };
      }

    my $gateway = ActiveMap::Utils::get_gateway_machine($args->{state}{target}) if $^O =~ /linux/;

    if (! $args->{pkt}) 	# initial call
      {

#	my @expected = ( $policies{'BSD'}, $policies{'first'}, $policies{'last'},
#			 $policies{'BSD-right'} );
	my @expected = values %policies;
        map { s/([0-9])/$1 x 8/ge } @expected;
	foreach my $data_str (@expected)
	  {
	    my $ip_id = int (rand (2**16));

	# data value => offset in 8-byte units, length
	    my @frags = ( 1 => 0, 3,  2 => 4, 2,  3 => 6, 3,  4 => 1, 4,
			  5 => 6, 3,  6 => 9, 3 );

	    my $totallen = 0;
	    while (@frags) {
	      my $data = shift @frags;
	      my $offset = shift @frags;
	      my $len = shift @frags;
	      $totallen += $len;

	      my $lastone = ! scalar(@frags);

	      my $frag = new Net::RawIP
		({
		  ip   => { saddr => ActiveMap::Utils::this_ip(), daddr => $args->{state}{target},
			    # MF = 13th bit, + 1 for icmp header
			    frag_off => ($lastone ? 0x0 : 0x2000) + $offset + 1,
			    protocol => 1,
			    tos => 0,
			    id => $ip_id,
			  },
		  generic => { data => $data x ($len * 8) }
		 });

	      if ($^O !~ /linux/)
		{
		  $frag->send();
		}
	      else
		{
		  $frag->ethnew(ActiveMap::Utils::get_pcap_device());
		  $frag->ethset(dest => $gateway);
		  $frag->ethsend();
		}
#	      select undef, undef, undef, 0.01;
	    }

	    # send the initial packet with ICMP header
	    # Have to send directly on ethernet since on Linux anyway it won't sent fragments

	    my $cksum = ActiveMap::Utils::inet_cksum
	      ( pack ("C2S3", 8, 0, 0, $args->{state}{icmp_id}, 0) . # ICMP header
		$data_str );
#	    print "checksum is $cksum";
	    my $initial = new Net::RawIP
	      ({
		ip => {  saddr => ActiveMap::Utils::this_ip(), daddr => $args->{state}{target},
			 frag_off => 0x2000, # offset = 0, set MF bit
			 id => $ip_id,
			 tos => 0,
			 protocol => 1
		      },
		icmp => { type => 8, code => 0, id => $args->{state}{icmp_id},
			  check => $cksum }
	       });
	    if ($^O !~ /linux/)
	      {
		$initial->send();
	      }
	    else
	      {
		$initial->ethnew(ActiveMap::Utils::get_pcap_device());
		$initial->ethset(dest => $gateway);
		$initial->ethsend();
	      }
	  }

	return;
      }

    # parse the reply to see which policy it was

    my $reply = $args->{pkt};
    my ($data) = $reply->get( { icmp => ['data'] } );

    $data =~ s/([0-9])\1{7}/\1/g;

    foreach my $policy (keys %policies)
      {
	my $len = length $policies{$policy};
	my $had_extra_data = ($len != length $data);
	$data =~ s/[^0-9]//g if $had_extra_data;
	$data = substr($data, 0, $len) if ($len != length $data);
	if ($policies{$policy} eq $data)
	  {
	    $args->{state}{result_ref}{overlap_policy} = 
	      $policy . ($had_extra_data ? '(also sent extra data)' : '');
	    $args->{done_flag} = 1;
	    return;
	  }
      }
      
    $args->{state}{result_ref}{overlap_policy} = "Problem: Returned '$data' but this is not a known policy!";
    $args->{done_flag} = 1;
    return;
  }


#
# Driver routines
#

sub _demux_packets
{
    my ($userdata, $hdr, $pkt) = @_;

    return if (!defined $hdr || !defined $pkt);

    # Dispatch based on packet contents; fn per test to do this?
    # Maybe just use echo reply seq # for now

    my $reply = new Net::RawIP({icmp => { type => 0 }});
    $reply->bset( $pkt, 1 );

    my ($icmp_id) = $reply->get( { icmp => [qw (id)] } );

#    print "Got a packet with icmp id $icmp_id\n";

    my $test;
    return if ! defined ($test = $id_map{$icmp_id});

    &{$ip_tests{$test->{testname}}{control_fn}} ({
		       state => $test,
		       timeout => 0,
		       hdr => $hdr,
		       pkt => $reply
		      });
}

#
# start test $test_name, update %test_results, %active_tests appropriately
#
# Returns 1 if started ok, 0 if no more iterations left
sub _start_test
  {
    my $test_name = shift;
    my $target = shift;

    my $test = $active_tests{$target.$test_name};
    if (defined $test)
      {
	return 0 if $test->{iters_left} == 0;
	push @{$results{$target}{$test_name}}, {};
	$active_tests{$target.$test_name} = {
		 icmp_id => get_next_icmp_id(),
		 time => time(),
		 iters_left => $test->{iters_left} - 1,
		 result_ref =>  @{$results{$target}{$test_name}}[-1],
		 target => $target,
		 testname => $test_name
		}
      }
    else
      {
	$results{$target}{$test_name} = [{}];
	$active_tests{$target.$test_name} = 
	  {
	   icmp_id => get_next_icmp_id(),
	   time => time(),
	   iters_left => $ip_tests{$test_name}{repeat_count} - 1,
	   result_ref => $results{$target}{$test_name}->[0],
	   target => $target,
	   testname => $test_name
	  };
      }

    $id_map{$active_tests{$target.$test_name}{icmp_id}} = $active_tests{$target.$test_name};

    &{$ip_tests{$test_name}{control_fn}}({ state => $active_tests{$target.$test_name},
					   timeout => 0,
					 });

    return 1;
  }

sub do_ip
{
    my ($max_hosts, $num_active_tests, $results, @targets) = @_;

   my $this_ip = ActiveMap::Utils::this_ip();

   %results = %{$results};

    my $pcap = ActiveMap::Utils::init_pcap(); 

    my @echo_capable;
    foreach (@targets)
      {
	if (! $results{$_}{icmp_svcs}{echo})
	  {
	    print "# Can't do IP mapping on $_: ICMP echo not enabled\n";
	  }
	else
	  {
	    push @echo_capable, $_;
	  }
      }
    @targets = @echo_capable;


    # We're using ICMP echo to do the IP tests; use pcap to grab replies
    my $filter_str = "dst host $this_ip and icmp and icmp[0]==0";
    foreach (keys %ip_tests)
      {
	my $subfilter = $ip_tests{$_}{filter};
	next if ! defined $subfilter;
	$filter_str .= " or ($subfilter)"
      }

    ActiveMap::Utils::set_pcap_filter( $pcap, $filter_str );

    $results{$_} = {} foreach (grep { ! exists $results{$_}  } (@targets)) ;

    my $max_at_once = int((ActiveMap::Firewall::num_ports_allocated() / 
			   $num_active_tests) / 2);

    $max_at_once = $max_hosts if $max_hosts < $max_at_once;

    while (@targets)
      {
	my @chunk_o_hosts = splice @targets, 0, $max_at_once;
	_do_ip_some_targets($pcap, $num_active_tests, @chunk_o_hosts);
      }

    Net::Pcap::close($pcap);

    return \%results;
}

sub _do_ip_some_targets
  {
    my ($pcap, $num_active_tests, @targets) = @_;
    my $this_ip = ActiveMap::Utils::this_ip();

    my @tests_to_run = keys %ip_tests;

    my %next_host_test_idx; # idx of test to run next for the given host

    # load up some initial tests
    $num_active_tests = scalar(@tests_to_run) if ($num_active_tests > scalar(@tests_to_run));

    foreach my $target (@targets) {
      for (my $i = 0; $i < $num_active_tests; $i++) {
	_start_test ($tests_to_run[$i], $target);
      }
      $next_host_test_idx{$target} = $num_active_tests;
    }

    # Make sure read timeout is something reasonable !!!

    while (scalar (keys %active_tests)) {
      Net::Pcap::dispatch( $pcap, -1, \&_demux_packets, \%active_tests );

      my $time = time();
      foreach my $test (keys %active_tests)
	{
	  my $testname = $active_tests{$test}{testname};  
	  
	  if (! $active_tests{$test}{done_flag} && $time > $active_tests{$test}{time} + $ip_tests{$testname}{test_timeout})
	      {
		&{$ip_tests{$testname}{control_fn}}({
					   result => $active_tests{$test}{result_ref},
					   state => $active_tests{$test},
					   timeout => 1,
					   hdr => undef, pkt => undef
					  });
		$active_tests{$test}{done_flag} = 1;
	      }
	    if ($active_tests{$test}{done_flag} == 1)
	      {
		my $target = $active_tests{$test}{target};
		if ( ! _start_test ($testname, $target) )
		  {
		    delete $active_tests{$test};
		    my $new_testname = $tests_to_run[$next_host_test_idx{$target}++];
		    _start_test ($new_testname, $target)
		      if defined $new_testname;
		  }
	      }
	  }
      }

    return \%results;
}



1;
