# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

#
# Measure network properties (topology, MTU, etc.) to end host
#

package ActiveMap::Network;

use strict;

use Net::RawIP;
use Net::Pcap;
use ActiveMap::Utils;
use ActiveMap::TCPSession;

#
# My globals 
#

#  {  host name/IP supplied => 
#             { hops => ,
#               _expected_hops => , # a guess value to help the search
#               mtu => ,
#               icmp_svcs => { echo => 1, ...],
#               udp_svcs => { dns|netbios => port, ...},
#               tcp_svcs => { http|ssh|ftp|telnet => port, ... } }
#             }
#  }
my %results;

sub results { return %results }

sub MTU_HOP_TIMEOUT { 10 }
sub ICMP_SVC_TIMEOUT { 3 }
sub TCP_SVC_TIMEOUT { 3 }


sub set_expected_hops
  {
    my ($host, $ttl) = @_;
    my $expected_hops;
    ($ttl <= 32 and $expected_hops = 33 - $ttl) ||
      ($ttl <= 64 and $expected_hops = 65 - $ttl) ||
	($ttl <= 128 and $expected_hops = 129 - $ttl) ||
	  ($expected_hops = 255 - $ttl);
    $expected_hops++; # sometimes it's off by one, who knows why?

    $results{$host}{_expected_hops} = $expected_hops;
#    $results{$host}{_expected_hops} = 64;
  }


sub demux_mtu_and_hops
  {
    my ($maps, $hdr, $pkt) = @_;

    my ($vals, $icmp_map, $tcp_map, $do_mtu) = @{$maps};

    my $proto = ActiveMap::Utils::get_pkt_proto($pkt, 1);
    if ($proto == 1) # icmp
      {
	my $reply = new Net::RawIP({icmp => {}});
	$reply->bset ($pkt, 1);
	my ($type, $icmp_id) = $reply->get ( { icmp => [qw(type id)] } );
#	print "echo reply with id $icmp_id\n";
	return if $type != 0 || ! defined $icmp_map->{$icmp_id};

#	print "echo reply from $icmp_map->{$icmp_id}\n";

	my $entry = $vals->{$icmp_map->{$icmp_id}};
	$entry->{low} = $entry->{try} if $do_mtu;
	$entry->{high} = $entry->{try} if ! $do_mtu;
	$entry->{found} = 1;
      }
    elsif ($proto == 6) # tcp
      {
	my $reply = new Net::RawIP({tcp=>{}});
	$reply->bset($pkt, 1);
	my ($dport, $ack_seq) = $reply->get ( { tcp => ['dest', 'ack_seq'] } );
	my $map_entry;
	return if ! defined ($map_entry = $tcp_map->{$dport});
	my $session = $map_entry->[1];
	return if ($session->{iss} + 1) != $ack_seq;
	my $entry = $vals->{$map_entry->[0]};
	$entry->{low} = $entry->{try} if $do_mtu;
	$entry->{high} = $entry->{try} if ! $do_mtu;
	$entry->{found} = 1;
      }
    else { die "weird packet\n" };
    # *** else why did we get this?
  }


# args = hosts
# *** NOTE: Only supports MTU up to 1500 since send() fails after that
# and never resets itself.
sub find_mtu_or_hops
  {
    my $do_mtu = shift;
    my @hosts = @_;

    # For MTU: high = this value does not work, low = this value does work,
    # For hops: high = this value does work, low = this value does not work
    # try = try this value next
    my %vals;

    %vals = map { $_ => { high => 1502, low => 300, try => 1500, found => 0 } } @hosts
      if $do_mtu;

    # since try is expected to be close to 0, we let it do its thing normally
    %vals = map { $_ => { high => 255,
			  low => 0,
			  try => $results{$_}{_expected_hops}, found => 0 } } @hosts
      if !$do_mtu;

    $results{$_}{hops} = 'Unknown: did not complete' foreach @hosts;

    my $this_ip = ActiveMap::Utils::this_ip();
    my $pcap = ActiveMap::Utils::init_pcap({ maxbytes => 100 });

    # look for echo replies and SYN/ACKs
    my $filter = "dst host $this_ip and ".
      "((icmp and icmp[0]==0) or " . 
	ActiveMap::Firewall::get_filter_str(). ")"; 

    ActiveMap::Utils::set_pcap_filter($pcap, $filter);

    my $start = time();

    # send out queries to the appropriate services
    while (scalar(keys %vals) && time() - $start < MTU_HOP_TIMEOUT) 
      {
	# icmp_id => host name being mapped
	my %icmp_map;

	# tcp port => [host, session]
	my %tcp_map;

	my $icmp_id = int(rand (2**15));

	# send packets out
	foreach my $host (keys %vals)
	  {
	    if ($results{$host}{icmp_svcs}{echo})
	      {
		my @icmp_addon = $do_mtu ? 
		  (data => 'A' x ($vals{$host}{try} - 28)) : ();
		my @ip_addon = $do_mtu ? (frag_off => 2**14) : 
		  (ttl => $vals{$host}{try});
		# eval because we get an error if > local MTU
		eval {
		  my $echo_request = new Net::RawIP
		    ({
		      ip => { saddr => $this_ip, daddr => $host,
			    @ip_addon},
		      icmp => { type => 8, code => 0, id => $icmp_id,
				@icmp_addon }
		     });
		  $echo_request->send();
		};

#		print "tried to send $vals{$host}{try} ttl to $host".
#		  " with id $icmp_id\n" if !$do_mtu;
#		print "$@\n" if $@;

		$icmp_map{$icmp_id} = $host;
		++$icmp_id;
	      }
	    else
	      {
		my @tcp_addon = $do_mtu ? 
		  (syndata => 'A' x ($vals{$host}{try} - 40), df => 1) :
		    (ttl => $vals{$host}{try});

		my $tcp_port = (values %{$results{$host}{tcp_svcs}})[0];
		if (!defined $tcp_port)
		  {
		    warn "# No relevant services found for $host; can't determine MTU\n";
		    delete $vals{$host};
		  }
		my $session = new ActiveMap::TCPSession
		  ({
		    dstip => $host, dstport => $tcp_port,
		    @tcp_addon
		   });

		eval # In case > local MTU 
		  {
		    $session->connect();
		  };
		$tcp_map{$session->{srcport}} = [$host, $session];
	      }
	  }


	my $localstart = time();
	DISPATCH: while (time() < $localstart+2) {
	  Net::Pcap::dispatch($pcap, -1, \&demux_mtu_and_hops, 
			      [\%vals, \%icmp_map, \%tcp_map, $do_mtu]);

	  foreach (values %vals)
	    {
	      next DISPATCH if ! $_->{found};
	    }
	  last; # no need to loop if everyone's done
	}

	# handle the case where no response packet gotten
	foreach my $host (keys %vals)
	  {
	   if (! $vals{$host}{found})
	     {
	       $vals{$host}{high} = $vals{$host}{try} if $do_mtu;
	       $vals{$host}{low} = $vals{$host}{try} if ! $do_mtu;
	     }

	   if ($vals{$host}{high} - $vals{$host}{low} == 1)
	     {
	       $results{$host}{mtu} = $vals{$host}{low} if $do_mtu;
	       $results{$host}{hops} = $vals{$host}{high} if ! $do_mtu;
	       delete $vals{$host};
	     }
	   else
	     {
	       if (! $do_mtu && $results{$host}{_expected_hops} - $vals{$host}{try} < 5)
		 {
		   --$vals{$host}{try};
		 }
	       else
		 {
		   $vals{$host}{try} = int(($vals{$host}{high} + $vals{$host}{low})/2);
		 }
	       $vals{$host}{found} = 0;
	     }
	  }

	# teardown any tcp sessions started
	foreach (values %tcp_map)
	  {
	    $_->[1]->teardown();
	  }

      }

    Net::Pcap::close($pcap);
  }


sub demux_echo
{
  my ($icmp_map, $hdr, $pkt) = @_;
  my $reply = new Net::RawIP({icmp => {}});
  $reply->bset( $pkt, 1 );
  my ($ttl, $icmp_id) = $reply->get( { ip => ['ttl'], icmp => [qw (id)] } );

  return if ! defined $icmp_map->{$icmp_id};
  $results{$icmp_map->{$icmp_id}}{icmp_svcs}{echo} = 1;

  return if defined $results{$icmp_map->{$icmp_id}}{_expected_hops};

  set_expected_hops($icmp_map->{$icmp_id}, $ttl);
}

sub test_icmp_echo
  {
    my ($pcap, $this_ip, @hosts) = @_;

    # look for echo replies
    my $filter = "dst host $this_ip and icmp and icmp[0]==0";
    ActiveMap::Utils::set_pcap_filter( $pcap, $filter );

    my %icmp_map = ();

    my $icmp_id = int(rand (2**15));
    foreach my $host (@hosts) {
      my $request = new Net::RawIP
	({
	  ip => { saddr => $this_ip, daddr => $host },
	  icmp => { type => 8, code => 0, id => $icmp_id }
	 });

      $request->send();
      $request->send(); # why not once more

      $icmp_map{$icmp_id} = $host;
      ++$icmp_id;
    }

    # wait for at most 2 seconds to see what comes back
    my $icmp_start = time();

    while (time() < $icmp_start + ICMP_SVC_TIMEOUT) {
      Net::Pcap::dispatch($pcap, -1, \&demux_echo, \%icmp_map);
    }

  }


my %tcp_port_map = ( 80 => 'http', 22 => 'ssh', 21 => 'ftp', 23 => 'telnet' );

sub demux_tcp
  {
    my ($tcp_map, $hdr, $pkt) = @_;
    my $reply = new Net::RawIP({tcp => {}});
    $reply->bset( $pkt, 1 );
    my ($ttl, $sport, $dport) = $reply->get ( { ip => ['ttl'], tcp => ['source', 'dest'] } );

    # *** Eventually look at data to figure out which protocol

    my ($svc, $host, $session);
    $svc = $tcp_port_map{$sport} or return;
    return if ! defined $tcp_map->{$dport};

    ($host, $session) = @{$tcp_map->{$dport}};
    $session->incoming_packet($reply);
    return if $session->{state} ne 'ESTABLISHED';

    $results{$host}{tcp_svcs}{$svc} = $sport;

    return if defined $results{$host}{_expected_hops};
    set_expected_hops($host, $ttl);
  }

sub test_tcp
  {
    my ($pcap, $this_ip, @hosts) = @_;

    my $filter = "dst host $this_ip and " .
      ActiveMap::Firewall::get_filter_str();

    ActiveMap::Utils::set_pcap_filter($pcap, $filter);

    # local port => host connected to, session
    my %tcp_map;

    foreach my $host (@hosts)
      {
	foreach my $port (keys %tcp_port_map)
	  {
	    my $session = new ActiveMap::TCPSession
	      ({ dstip => $host, dstport => $port });
	    $session->connect();
	    $tcp_map{$session->{srcport}} = [$host, $session];
	  }
      }

    # wait for 2-3 seconds to see what comes back
    my $start = time();

    while (time() < $start + TCP_SVC_TIMEOUT) {
      Net::Pcap::dispatch($pcap, -1, \&demux_tcp, \%tcp_map);
    }

    foreach (values %tcp_map) {
      $_->[1]->teardown();
    }
  }

# Right now searches for (but doesn't even verify the protocol on):
# ICMP echo
# HTTP on port 80
# SSH on port 22
# FTP on port 21
# Telnet on port 23
sub find_svcs
  {
    my @hosts = @_;

    my $this_ip = ActiveMap::Utils::this_ip();
    my $pcap = ActiveMap::Utils::init_pcap({ maxbytes => 100 }); # only need headers

    test_icmp_echo($pcap, $this_ip, @hosts);
    test_tcp($pcap, $this_ip, @hosts);

    Net::Pcap::close($pcap);
  }

# Returns results hashref
sub do_network
  {
    my @hosts = @_;

    my $max_at_once = int (ActiveMap::Firewall::num_ports_allocated() /
			   (scalar (keys %tcp_port_map) + 2));

    $results{$_} = { icmp_svcs => {}, tcp_svcs => {}, udp_svcs => {}  }
      foreach (@hosts);

    while (@hosts)
      {
	my @chunk_o_hosts = splice @hosts, 0, $max_at_once;
	warn "# Doing service discovery at time " . time() . "\n";
	find_svcs(@chunk_o_hosts);
	warn "# Finding MTU at time " . time() . "\n";
	find_mtu_or_hops(1, @chunk_o_hosts);
	warn "# Finding hop count at time " . time() . "\n";
	find_mtu_or_hops(0, @chunk_o_hosts);
      }

    return \%results;
  }


1;
