# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

#
# Test for TCP behavior
#

package ActiveMap::TCP;

use strict;
use ActiveMap::Utils;
use Net::RawIP;

#
# Globals
#

#
# test name => { control_fn => \&fn, repeat_count => , test_timeout => , 
#                svcs => ['http', ...] list of preferred svcs, can be []
#               (optional) autoack => 1, always ACK data sent to us
#               (optional)  synoptions => { optid => optdata, ... },
#               (optional)  synflags => { rst => 1, ... }, syndata => $data_with_SYN_packet
#

my $unhandled = 0;

my %set1 = 
  (
       rst_in_seq => {
    		     control_fn => sub { rst_test(0, @_) }, 
    		     repeat_count => 3,
    		     test_timeout => 8,
    		     svcs => ['http', 'ftp', 'telnet', 'ssh']
    		    },

       rst_in_window => {
    		     control_fn => sub { rst_test(1, @_) },
    		     repeat_count => 3,
    		     test_timeout => 8,
    		     svcs => ['http', 'ftp', 'telnet', 'ssh']
    		    },

       rst_outside_window => {
    		     control_fn => sub { rst_test(-1, @_) },
    		     repeat_count => 3,
    		     test_timeout => 8,
    		     svcs => ['http', 'ftp', 'telnet', 'ssh']
    		    },
);

my %set2 =
(
   overlapping_seg => {
  		       control_fn => \&overlapping_seg,
  		       repeat_count => 3,
  		       test_timeout => 6, # phase 1 timeout
  		       svcs => ['http', 'ssh'],
		       autoack => 1
     		      },

  
);

my %tcp_tests = (%set1, %set2);

# hostname => { name of test => array of test results (one per iteration) }
# (each result is a hashref)
my %results;

# targettestname
#  {
#    testname => name of test
#    timeout => time test should time out
#    iters_left => num. remaining iterations,
#    result_ref => \%curr_result,
#    session => tcp_session,
#    done_flag => true if test should be ended (result already obtained)
#    control_fn => copied from tcp_tests,
#    phase => internal state variable, starts at 0
#    target => target host
#    target_port => port
#    target_svc => 'http' or 'ssh' ...
#    autoack => (from tcp_tests)
#  }
my %active_tests;
my %port_map; # also tcp source port => active_test

#
#  Implementation of TCP tests
#

#
# $args = { state => $active_tests{target.testname},
#           timeout => 1 if this test timed out -- results should be written
#           pktobj => Net::RawIP object }
# Also given ($data, $rst, $fin)

# send_at_offset = 0 for in sequence, 1 for in window, -1 for outside window
sub rst_test
  {
   my $send_at_offset = shift;
   my $args = shift;
   my ($data, $rst, $fin) = @_;

   my $session = $args->{state}{session};

   my %key = ( 0 => 'accepts_rst_in_sequence',
	       1 => 'accepts_rst_in_window',
	       -1 => 'accepts_rst_outside_window' );
   $args->{state}{key} = $key{$send_at_offset}
     if ! defined $args->{state}{key};

   $send_at_offset = ($session->{rcv_wnd} + 33322) if $send_at_offset < 0;

   if ($args->{timeout})
     {
       $args->{state}{result_ref}{$args->{state}{key}} = 
	 $args->{state}{sent_rst} ? 'yes' : 'did not complete';
       $args->{state}{done_flag} = 1;
#       print "rst_test with offset $send_at_offset timed out at ".time()." for $args->{target}\n";
       return;
     }

   if (! defined $args->{pktobj})
     {
       # initial state; connect
       warn "Control fn called w/o packet in state other than START"
	 if $session->{state} ne 'START';
       $args->{state}{sent_rst} = 0;
       $session->connect();
       return;
     }

#   print "Got '$data' in state $args->{state}{session}{state}\n";
   return if $args->{state}{session}{state} ne 'ESTABLISHED';

   if ( ! $args->{state}{sent_rst} )
     {
       # Send RST before ACK'ing the SYN/ACK so it doesn't start sending us data
       $session->send_packet('', ['rst'],
			     undef, # opts
			     $send_at_offset ); # seq num

#       print "sending RST with seq number " . ($session->{snd_nxt} + $session->{snd_wnd}+33322) . "\n";

       $session->send_packet('', ['fin']); # to provoke an ACK if the RST not accepted
       $args->{state}{sent_rst} = 1;

       return;
     }

   if ($rst)
     {
       $args->{state}{result_ref}{$args->{state}{key}} = 'yes';
#       print "does accept rst outside\n";
       $args->{state}{done_flag} = 1;
     }
   else
     {
       $args->{state}{result_ref}{$args->{state}{key}} = 'no';
#       print "does not accept rst outside\n";
       $args->{state}{done_flag} = 1;
     }
  }


#
# Send the following segment sets
#
# For SSH:
# S|SH-|          For SSH: Send 'SH-', then ' 2.0-blah'; first and BSD will work, last won't
#    |X2.0-blah\r\n|  (Initial 'S' last) 
#
# S|SH|  |+|        In this case, first and last will not work but BSD will
#   |X-2.0-blah\r\n|
#
# Not working is basically defined as the reception of a FIN and/or RST packet
#
# Anything else is new and should be investigated.
#
# For HTTP:
# GET| /| HTTP/1.0\r\n\n|  First, BSD work
#     |! |
#
# GET| /| X| HTTP/1.0\r\n\n|  First, last won't work, BSD will
#     |!   |
#
sub overlapping_seg
  {
    my $args = shift;
    my ($data, $rst, $fin) = @_;

    my $state = $args->{state};

    my $msg = sub {};
#    eval ('$msg = sub {
 #     print "# TCPoverlap $state->{target}, local port $state->{session}{srcport}, phase $state->{phase}, round $state->{round}  :  $_[0]\n";
  #  }');

     &$msg ("data gotten: $data") if $data ne '';
    if ($args->{timeout} && $state->{phase} eq 'connect')
      {
	if (! keys %{$state->{result_ref}}) {
	&$msg ('could not connect');
	  $state->{result_ref}{tcp_segment_overlap} = 'No result: Could not connect';
	}
	$state->{done_flag} = 1;
	return;
      }

    if ($args->{timeout} && $state->{round} >= 2) # because 2 rounds
      {
	$state->{round2} = 1 if ! defined $state->{round2};

	if (! keys %{$state->{result_ref}}) {
	  $state->{result_ref}{tcp_segment_overlap} = undef;
	  my $result = \$state->{result_ref}{tcp_segment_overlap};
	  $$result = 'last' if ($state->{round1} == 0 && $state->{round2} == 0);
	  $$result = 'first' if ($state->{round1} == 1 && $state->{round2} == 0);
	  $$result = 'BSD' if ($state->{round1} == 1 && $state->{round2} == 1);
	  $$result = 'strange' if ($state->{round1} == 0 && $state->{round2} == 1);
	}
	$state->{done_flag} = 1;
	return;
      }

    if (! $state->{round} && ! defined $args->{pktobj}) 	# initial call
      {
	$state->{session}->connect();
	$state->{phase} = 'connect';
	$state->{round} = 1;
	$state->{round1} = undef;
	$state->{round2} = undef;
	return;
      }

    &$msg ("received data '$data', phase is $state->{phase}") if $data ne '';

    if ($state->{phase} eq 'connect') {
      return if $state->{session}{state} ne 'ESTABLISHED';

      if ( $state->{target_svc} eq 'ssh' )
	{
	  # make sure we got a nice welcome from the SSH server ('SSH-d.dd-comments')
	  if ( $data =~ /^\s*SSH-([0-9][0-9.]*)-/) 
	    {
	      $state->{ssh_version} = $1;
	      &$msg ("ssh version $1\n");
	      &$msg ( "Saw SSH string, sending segments");
	    }
	  else
	    {
	      &$msg ("Received data other than version string") if $data ne '';
	      return; # wait for the server's version string to arrive
	    }
	}

      # offset from seq point => data to send
      my @round1;
      my @round2;
      if ( $state->{target_svc} eq 'ssh' )
	{
	  @round1 = ( 1 => 'SH-', 
		      3 => "X$state->{ssh_version}-round1\r\n", 
		      0 => 'S' );
	  @round2 = ( 1 => 'SH',  
		      7 => '+', 
		      2 => "X-$state->{ssh_version}-round2\r\n", 
		      0 => 'S' );
	}
      else
	{
	  @round1 = ( 3 => ' /',
		      4 => '! ',
		      5 => " HTTP/1.0\r\n\n",
		      0 => 'GET' );
	  @round2 = ( 3 => ' /',
		      4 => '!   ',
		      5 => ' X',
		      7 => " HTTP/1.0\r\n\n",
		      0 => 'GET' );
	}

      my @frags;
      @frags = $state->{round} == 1 ? @round1 : @round2;
      my $snd_nxt;
      while (@frags) {
	my $offset = shift @frags;
	my $data = shift @frags;

	my $end = $state->{session}{snd_nxt} + $offset + length($data);
	$snd_nxt = $end if $end > $snd_nxt;

	$state->{session}->send_packet($data, [], undef, $offset);
      }
      $state->{session}{snd_nxt} = $snd_nxt; # session gets confused otherwise

      $state->{phase} = 'response';
      return;
    }

    # deal with a response packet
    my $round_result = $state->{round} == 1 ? 
      \$state->{round1} :
      \$state->{round2};

#    warn "# Tgt svc is $state->{target_svc}\n";
    if ( $state->{target_svc} eq 'http' )
      {
	if ( $args->{timeout} || $rst )
	  {
	    $state->{result_ref}{tcp_segment_overlap} = 'No result: No response from HTTP server';
	    $state->{done_flag} = 1;
	    return;
	  }
	else # parse response
	  {
	    my ($resp) = ($data =~ /(HTTP.*?\n)/);
	    return if (!$resp);
#	    warn "# HTTP Response $resp\n";
	    $$round_result = ($resp =~ /400/) ? 0 : 1;
	  }
      }
    else # ssh
      {
	if ($args->{timeout})
	  {
	    &$msg ( "round 1 timed out, declaring success");
	    $$round_result = 1;
	  }

	if ($fin || $rst) {
	  $$round_result = 0; # "failed"
	  &$msg ( "got a fin or rst in round $state->{round} ");
	} elsif (! defined $$round_result) {
	  #      print "got a message: $data\n";
	  return;
	}
      }

    if ($state->{round} == 1)
      {
	&$msg ("round 1 done, result = $$round_result\n");
	delete $port_map{$state->{session}{srcport}};
	$state->{session}->teardown();
	# reconnect and begin round 2
	$state->{timeout} = time() + $tcp_tests{$state->{testname}}{test_timeout};
	$state->{phase} = 'connect';
	$state->{round} = 2;
	$state->{session} = new ActiveMap::TCPSession
	  ({ dstip => $state->{target},
	     dstport => $state->{target_port} });
#	print "connecting round 2 to port $state->{target}:$state->{target_port} at " . time() . "\n";
	$state->{session}->connect();
	$port_map{$state->{session}{srcport}} = $args->{state};
	return;
      }
    else
      {
	# force termination
	return overlapping_seg( { state => $args->{state}, timeout => 1 } );
      }
  }


#
# start test $test_name, update %results, %active_tests appropriately
#
# Returns 1 if started ok, 0 if no more iterations left
sub _start_test
  {
    my $test_name = shift;
    my $this_ip = shift;
    my $target = shift;

    die "Invalid test name: $test_name" if !defined $tcp_tests{$test_name};

#    warn "# starting test $test_name for $target at " . time() . "\n";
#    print "# caller is ", caller(), "\n";

    my %params;
    if (exists $active_tests{$target.$test_name})
      {
	return 0 if $active_tests{$target.$test_name}{iters_left} == 0;
	push @{$results{$target}{$test_name}}, {};
	my $entry = $active_tests{$target.$test_name};
	%params = (
		   iters_left => $entry->{iters_left} - 1,
		   result_ref => @{$results{$target}{$test_name}}[-1],
		   target_port => $entry->{target_port},
		   target_svc => $entry->{target_svc},
		  );
      }
    else
      {
	$results{$target}{$test_name} = [{}];

	# Choose a service according to the test's preferences

	my $target_port;
	my $target_svc;
	if (! scalar(@{$tcp_tests{$test_name}{svcs}}))
	    {
	      $target_svc = (keys %{$results{$target}{tcp_svcs}})[0];
	      $target_port = $results{$target}{tcp_svcs}{$target_svc};
	    }
	else
	  {
	    foreach my $svc (@{$tcp_tests{$test_name}{svcs}})
	      {
		if (defined $results{$target}{tcp_svcs}{$svc})
		  {
		    $target_port = $results{$target}{tcp_svcs}{$svc};
		    $target_svc = $svc;
		    last;
		  }
	      }
	    if (!$target_port)
	      {
		$results{$target}{$test_name}->[0] = 
		  'No result: no appropriate service found, needs one of (' .
		    ( join (' ', @{$tcp_tests{$test_name}{svcs}}) ) . ')';
		$active_tests{$target.$test_name} = 
		  { done_flag => 1, testname => $test_name, target => $target,
		    iters_left => 0 };
		return 0;
	      }
	  }

	%params =
	  (
	   iters_left => $tcp_tests{$test_name}{repeat_count} - 1,
	   result_ref => $results{$target}{$test_name}->[0],
	   target_port => $target_port,
	   target_svc => $target_svc,
	  );
      }

    $active_tests{$target.$test_name} =  
      {
       timeout => time() + $tcp_tests{$test_name}{test_timeout},
       session => new ActiveMap::TCPSession({ dstip => $target, 
					      dstport => $params{target_port},
					      synoptions => $tcp_tests{$test_name}{synoptions},
					      synflags => $tcp_tests{$test_name}{synflags},
					      syndata => $tcp_tests{$test_name}{syndata}
					    }),
       done_flag => 0,
       control_fn => $tcp_tests{$test_name}{control_fn},
       phase => 0,
       target => $target,
       testname => $test_name,
       autoack => $tcp_tests{$test_name}{autoack},
       %params 
      };
    
    &{$tcp_tests{$test_name}{control_fn}}({
					     this_ip => $this_ip,
					     target => $target,
					     state => $active_tests{$target.$test_name},
					     pktobj => undef,
					     timeout => 0
					    });

    $port_map{$active_tests{$target.$test_name}{session}{srcport}} = 
      $active_tests{$target.$test_name};

#    print "test $target.$test_name uses src port ".
#      $active_tests{$target.$test_name}{session}{srcport} . "\n";

    return 1;
  }


sub _demux_packets
  {
    # call appropriate session->incoming_packet
    my ($userdata, $hdr, $pkt) = @_;

    return if (!defined $hdr || !defined $pkt);

    my $reply = new Net::RawIP({tcp => {}});
    $reply->bset( $pkt, 1 );

    my ( $tot_len ) = $reply->get( { ip => ['tot_len'] } );

    ### Sometimes the length is wrong -- see if this happens elsewhere!
#          if ($tot_len != length($pkt) ) # ***
#            {
#  	    print STDERR "#Fixing length: ip len is $tot_len, actual length is ". length($pkt). "\n";
#  	    substr ($pkt, $tot_len) = '';
#  	    $reply->bset( $pkt, 1 );
#            }

    my ( $saddr, $daddr, $srcport, $dstport ) = $reply->get( { ip => ['saddr', 'daddr'], 
					       tcp => ['source', 'dest'] } );

#    my  ( $dstport ) = $reply->get( { tcp => ['dest'] } );

#    print "src & dst: $saddr/$srcport  $daddr/$dstport\n";

    my $test = $port_map{$dstport};

      if (!defined $test) # ***
        {
#	my ($data) = $reply->get({tcp=>['data']});
#  	print "did not have a test for port $dstport;threw away data $data\n";
        $unhandled++;
        }
#      else
#        {
#  	print "had a test for $dstport\n";
#        }

    return if (! defined $test);
    my $session = $test->{session};

    my ($data, $rst, $fin, $valid_ack) = $session->incoming_packet($reply);

#    print "# Data in demux : '$data'\n" if $data ne '';

    if (! $valid_ack)
      {
	# *** Maybe do something with this case later
	warn "invalid ack, could be from new connection\n";
	return;
      }

      if ($test->{autoack} && !$rst && !$fin)
        {
	  if ($test->{testname} eq 'overlapping_seg' && $data ne '')
	    {
#	      print "# Sending autoack for overlapping seg to $session->{dstip}, data='$data'\n";
	    }
  	$session->send_packet('');
        }

    # call control_fn
    &{$test->{control_fn}}( {state => $test, timeout => 0, pktobj => $reply},
			  $data, $rst, $fin);
  }

sub do_tcp
{
    my ($max_hosts, $num_active_tests, $results, @hosts) = @_;
    $num_active_tests = scalar(keys %tcp_tests) 
      if ($num_active_tests > scalar(keys %tcp_tests));

    %results = %{$results};


    my @have_tcp_svcs;
    foreach (@hosts)
      {
	if (! scalar (keys %{$results{$_}{tcp_svcs}}))
	  {
	    warn "# Not doing TCP mapping on $_ since no supported TCP services available.\n";
	  }
	else
	  {
	    push @have_tcp_svcs, $_;
	  }
      }
    @hosts = @have_tcp_svcs;

    my $pcap = ActiveMap::Utils::init_pcap({ maxbytes => 8000 });

    my $filter = ActiveMap::Firewall::get_filter_str();
    ActiveMap::Utils::set_pcap_filter( $pcap, $filter );

    my $max_at_once = int((ActiveMap::Firewall::num_ports_allocated() / 
			   ($num_active_tests + 1))); # + 1 for safety

    $max_at_once = $max_hosts if $max_at_once > $max_hosts;
#    warn "# Mapping at most $max_at_once hosts at a time\n";

    while (@hosts)
      {
	my @chunk_o_hosts = splice @hosts, 0, $max_at_once;
	_do_tcp_some_targets($pcap, $num_active_tests, [sort keys %set1], @chunk_o_hosts);
	_do_tcp_some_targets($pcap, $num_active_tests, [sort keys %set2], @chunk_o_hosts);
      }

    Net::Pcap::close($pcap);

    my %stats;
    Net::Pcap::stats($pcap, \%stats);
    warn "# (For TCP tests) Packets received/dropped by pcap/dropped by iface: ".
        "$stats{ps_recv} / $stats{ps_drop} / $stats{ps_ifdrop}\n";

    warn "# Number of packets received but were not handled: $unhandled\n";

    return \%results;
}

sub _do_tcp_some_targets
{
  my ($pcap, $num_active_tests, $test_set, @targets) = @_;

  my $this_ip = ActiveMap::Utils::this_ip();

  my @tests_to_run = @{$test_set};
  $num_active_tests = scalar(@tests_to_run) if scalar(@tests_to_run) < $num_active_tests;  

  my %next_host_test_idx; # idx of test to run next for the given host

  # load up some initial tests
  foreach my $target (@targets)
    {
      for (my $i = 0; $i < $num_active_tests; $i++)
	{
	  _start_test ($tests_to_run[$i], $this_ip, $target);
	}
      $next_host_test_idx{$target} = $num_active_tests;
    }

  my $maxpkts = 0;
  while (scalar (keys %active_tests))
    {
      my $numpkts = Net::Pcap::dispatch( $pcap, -1, \&_demux_packets, undef );
      if ($numpkts > $maxpkts)
	{
	  $maxpkts = $numpkts;
#	  print "# New max packets dispatched: $numpkts\n";
	}

      my $time = time();
      foreach my $test (keys %active_tests)
	{
	  my $testname = $active_tests{$test}{testname};
	  if (! $active_tests{$test}{done_flag} && $time > $active_tests{$test}{timeout})
	    {
	      &{$tcp_tests{$testname}{control_fn}}({
						    result => $active_tests{$test}{result_ref},
						    state => $active_tests{$test},
						    timeout => 1,
						    pktobj => undef
						   });
	      #	      print "#timing out test $testname for $active_tests{target} at " . time() . "\n";
	    }

	  if ($active_tests{$test}{done_flag} == 1)
	    {
	      # reset connection, release source port
	      if ($active_tests{$test}{session})
		{
		  delete $port_map{$active_tests{$test}{session}{srcport}};
		  $active_tests{$test}{session}->teardown() 
		    if ($active_tests{$test}{session});
		}
	      my $target = $active_tests{$test}{target};
	      if ( ! _start_test ($testname, $this_ip, $target) )
		{
		  delete $active_tests{$test};
		  my $new_testname = $tests_to_run[$next_host_test_idx{$target}++];
		  _start_test ($new_testname, $this_ip, $target)
		    if defined $new_testname;
		}
	      }
	}
    }

    return \%results;

}

1;
