# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

#
# ======== ActiveMap::TCPSession is a class for handling user-level TCP sessions ==========
#

package ActiveMap::TCPSession;

use strict;
use Net::RawIP;
use ActiveMap::Utils;
use ActiveMap::Firewall;

sub SYNTIMEOUT { 3 };  # *** 3 seconds is arbitrary; verify

#
# Session variables
# {/* conn target */ 
#  dstip, dstport, 
#
#       /* sender info, from RFC 793 */
#  iss     => initial sent sequence number,
#  snd_una => oldest unacknowledged sequence number,
#  snd_nxt => next sequence number to send
#  snd_wnd => sender window size, 
#  sndmss  => sender maximum segment size,
#        /* Receiver  info */
#  irs     => initial received sequence number;
#  rcv_wnd => receiver window size,
#  rcv_nxt => next sequence number to expect,
#  rcv_maxseq => maximum sequence number seen thus far
#  rcv_mss => receiver maximum segment size
#  ttl => IP ttl to set on packets
#  df => set DF flag on packets
#  _next_ip_id => IP id field counter
#
#  /* timing  */      Not doing this stuff -- Umesh
#  /* double rtt; */
#
# /* state */
# state => START | SYN (sent SYN) | ESTABLISHED (recvd SYN/ACK) | CLOSED (teardown done)
# no_rst => don't RST on DESTROY
#
# /* unread data buffer */
# databytes => how much of the data has been acked (no holes below this amount)
# data => all incoming data on the connection


#
# By default this is done in DESTROY, so be sure that this is what you want.
# To disable rst-on-destroy, set the 'no_teardown' flag
#


#
# Constructor expects
# {
#   dstip, dstport,
#   synoptions => { option_id => option_data  }, (optional)
#   mss, iptos (optional),
#   synflags => { fin => 1, ... (or whatever flags to pass with syn packet) }, (optional)
#   syndata => data to pass with syn packet (optional);
#   ttl => TTL to set on packets (optional)
#   df => 1 if DF flag to be set on packets (optional)
# }
#
sub new
  {
    my $CLASS = shift;
    my $args = shift;

    return bless {
		  srcip      => ActiveMap::Utils::this_ip(),
		  srcport    => undef, # set on connect
		  dstip      => $args->{dstip},
		  dstport    => $args->{dstport},

		  snd_wnd    => 5840,
		  snd_mss     => defined $args->{mss} ? $args->{mss} : 536,

		  rcv_mss    => undef,
		  rcv_wnd    => undef, # determined with SYN/ACK

		  synoptions => $args->{synoptions} || {},

		  synflags   => $args->{synflags} || {},
		  syndata    => defined $args->{syndata} ? $args->{syndata} : '',

		  ttl        => $args->{ttl} || 255,
		  df         => $args->{df} || 0,
		  _next_ip_id => int (rand (2**16)),

		  databytes  => 0,
		  data       => '',

		  state => 'START',
		 }, $CLASS;
  }


sub next_ip_id
  {
    my $self = shift;
    $self->{_next_ip_id}++;
    $self->{_next_ip_id} -= 2**16 if ($self->{_next_ip_id} >= 2**16);
    return $self->{_next_ip_id};
  }

#
# Takes no arguments (they've already been set)
# Returns which source port it got
sub connect
  {
    my $self = shift;

    my $iss = int (rand(2**30));
    $self->{snd_una} = $self->{snd_nxt} = $self->{iss} = $iss;
    $self->{srcport} = ActiveMap::Firewall::get_tcp_source_port();
    $self->{data} = '';
    $self->{databytes} = 0;

    my $syn_pkt = new Net::RawIP ( {
		    ip  => {
			    defined $self->{iptos} ? (tos => $self->{iptos}) : (),
			    saddr => ActiveMap::Utils::this_ip(),
			    daddr => $self->{dstip},
			    ttl => $self->{ttl},
			    frag_off => $self->{df} * 2**14,
			    id => $self->next_ip_id()
			   },
		    tcp => {
			    source => $self->{srcport},
			    dest => $self->{dstport},
			    seq => $self->{iss},
			    syn => 1,
			    data => $self->{syndata},
			    window => $self->{snd_wnd},
			    defined $self->{synflags} ? %{$self->{synflags}} : (),
			   },
		   } );

    my (@opt_type, @opt_data);
    foreach (keys %{$self->{synoptions}})
      {
	push @opt_type, $_;
	push @opt_data, $self->{synoptions}{$_};
      }

    # *** put back $syn_pkt->optset( tcp => { type => \@opt_type, data => \@opt_data  } );

    $syn_pkt->send();

    $self->{snd_nxt}++; # SYN counts as 1
    $self->{state} = 'SYN';

    return $self->{srcport};
  }

sub send_rst_back_to
  {
    my $rawip_pkt = shift;
    my ($srcip, $dstip, $srcport, $dstport, $ackseq) = 
      $rawip_pkt->get( { ip => [qw(saddr daddr)], tcp => [qw(source dest ack_seq)] } );

    # turn around ip addrs, ports, send rst
    my $rst = new Net::RawIP ( {
				ip => { saddr => $dstip, daddr => $srcip },
				tcp => { source => $dstport,
					 dest => $srcport,
					 seq => $ackseq,
					 rst => 1}
			       });
    $rst->send();
  }


#
# Only accepts data at the next sequence number (no holes permitted at any time)
# Returns ($data received, $rst, $fin, $valid_ack)
#
sub incoming_packet
  {
    my $self = shift;

    my $rawip_pkt = shift; # should be a Net::RawIP object

    if ($self->{state} eq 'START')
      {
	warn "# Received packet on port $self->{srcport} before SYN sent, ignoring...\n";
	return ('', 0, 0, 0);
      }

    my ($seq, $ack_seq, $ack, $rst, $syn, $fin, $window, $urg_ptr, $data) = 
      $rawip_pkt->get( { tcp => [qw(seq ack_seq ack rst syn fin window urg_ptr data)] } );

    if (! $rst && (! $ack || $ack_seq < $self->{snd_una} || $ack_seq > $self->{snd_nxt}))
      {
	warn "# Received a packet without the ACK bit set or\n".
 	  "# with invalid ACK sequence number, dropping. ack'ed $ack_seq, una = $self->{snd_una}, nxt = $self->{snd_nxt}\n";
#	_send_rst_back_to($rawip_pkt);
	return ('', 0, 0, 0);
      }

    if ($rst)
      {
	return ('', 1, 0, 1);
      }

    if ($self->{state} eq 'SYN') {
      if ($fin)
	{
	  warn "# Got FIN before SYN/ACK, dropping segment.\n";
	  return ('', 0, 0, 0);
	}
      if ($syn && $ack) {
	$self->{state} = 'ESTABLISHED';
	$self->{maxseqseen} = $self->{irs} = $seq;
	$self->{rcv_nxt} = $seq;
	warn " # SYN/ACK acknowledged unexpected sequence number" if $self->{iss} != $ack_seq - 1;
	$self->{snd_una} = $self->{iss} + 1;
	$self->{rcv_wnd} = $window;

	$data = ''; # strip off data from SYN packet

	my @opts = $rawip_pkt->optget( tcp => { type => [2] } );
	for (my $i = 0; $i < @opts; $i += 3) {
	  if ($opts[$i] == 2)	# MSS
	    {
	      $self->{snd_mss} = unpack "n", $opts[$i+2];
	      # our own MTU-based limit, assuming Ethernet 802.3 encapsulation
	      # see Stevens vol 1 p. 237
	      $self->{snd_mss} = 1500-48 if ($self->{snd_mss} > 1500-48);
	    }
	}

#	print "connection estd to $self->{dstip}:$self->{dstport}".
#	  "from $self->{srcport} with seq #$seq\n";

      } else {
	warn "# Expecting SYN/ACK, got non SYN/ACK packet, data=$data\n";
	return ('', 0, 0, 0);
      }
    }

    $self->{snd_una} = $ack_seq;

    if ($seq != $self->{rcv_nxt}) {
#      warn "# Ignoring out-of-order or duplicate or overlapping tcp segment: '$data'";
 #     print "# expected $self->{rcv_nxt}, got $seq\n";
      $self->{rcv_nxt} = $seq;
      return (0, 0, 0, 1);
    }

    if (length($data))
      {
	$self->{data} .= $data;
#	print "got data $data\n";
	$self->{rcv_nxt} = $seq + length($data);
      }
    else
      {
	++$self->{rcv_nxt} if $fin;
	++$self->{rcv_nxt} if $syn;
      }

    return ($data, 0, $fin, 1);
  }


# $flags = [ 'syn', 'urg', ... ]
# $opts = optional set of tcp options, [ opt1 option code, opt1 data, ... ]
# $seq_off = sequence number offset from snd_nxt to send at; 
#        if undef, send at expected next
# $ack_seq = what sequence number to ack; if undef, ack the last received seq number
# Returns number of segments (packets) sent.
sub send_packet
  {
    my $self = shift;
    my ($data, $flags, $opts, $seq_off, $ack_seq) = @_;

    my $num_segments = 0;

    do
      {
	my $datapiece = substr ($data, 0, $self->{snd_mss}, '');
	my $args = {
			      ip  => {
				      defined $self->{iptos} ? (tos => $self->{iptos}) : (),
				      saddr => ActiveMap::Utils::this_ip(),
				      daddr => $self->{dstip},
				      ttl => $self->{ttl},
				      frag_off => $self->{df} * 2**14,
				      id => $self->next_ip_id()
				     },
			      tcp => {
				      source => $self->{srcport},
				      dest => $self->{dstport},
				      seq => int(0 + $seq_off + $self->{snd_nxt}),
				      data => $datapiece,
				      window => $self->{snd_wnd},
				      ack => 1,
				      ack_seq => int($ack_seq || $self->{rcv_nxt}),
				      (map {$_ => 1} @{$flags})
				     }
			     };
	my $pkt = new Net::RawIP ($args);
	if ($opts)
	  {
	    my @opt_type;
	    my @opt_data;
	    while (@$opts)
	      {
		push @opt_type, shift @$opts;
		push @opt_data, shift @$opts;
	      }

	    $pkt->optset( tcp => { type => \@opt_type, data => \@opt_data  } );;
	  }

	$pkt->send();
	$self->{snd_nxt} += length($datapiece) if ! $seq_off;
	$self->{snd_nxt} += scalar(grep /fin/, @$flags) if ! $seq_off; # SYN, FIN incr seq num
	++$num_segments;
      } while (length $data);

    return $num_segments;
  }

# Send RST, relinquish source port
sub teardown
  {
    my $self = shift;
    return if ($self->{state} eq 'START' || $self->{state} eq 'CLOSED');
    $self->send_packet('', ['rst']);
    ActiveMap::Firewall::free_tcp_source_port($self->{srcport});
    $self->{state} = 'CLOSED';
  }

sub DESTROY
  {
    my $self = shift;
    $self->teardown();
  }

1;


