# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#
#
# Utility routines for packet management
#

package ActiveMap::Utils;

use Net::Netmask;
use Sys::Hostname;

#
# Pcap state globals
#
my $pcap_t;
my $pcap_dumper_t;
my $pcap_dev;
my $pcap_net;
my $pcap_mask;



#
# IMPORTANT NOTE: For proper operation on Linux, the version of pcap
# that comes with nmap must be installed. It includes a patch that
# lets read timeouts work properly.
#


# Can opt. pass in { dev => /dev/netdevice,  maxbytes => 1000, timeout => 1000ms }
# returns $pcap_t
sub init_pcap
{
    my $errbuf;
    $pcap_dev = $_[0]->{dev} || Net::Pcap::lookupdev(\$errbuf);
    die "Could not local pcap device: $errbuf\n" if ! defined $pcap_dev;

    my $maxbytes = $_[0]->{maxbytes} || 1000;
    my $timeout  = $_[0]->{timeout} || 1000;
    
    die $errbuf if (Net::Pcap::lookupnet($pcap_dev, \$pcap_net, \$pcap_mask,
                                             \$errbuf) == -1);

    $pcap_t = Net::Pcap::open_live($pcap_dev, $maxbytes,
                                          1,  # not promiscuous
                                          $timeout,
                                          \$errbuf);

    die $errbuf unless $pcap_t;

    return $pcap_t;
}


sub get_pcap_device { return $pcap_dev; }

sub set_pcap_filter
{
    my $pcap_t = shift;
    my $filter_str = shift;

    my $bpf_prog;
    die Net::Pcap::geterr($pcap_t)
        if (Net::Pcap::compile($pcap_t, \$bpf_prog, $filter_str, 0, $pcap_mask) == -1) ||
            (Net::Pcap::setfilter($pcap_t, $bpf_prog) == -1);
}

my $this_ip = undef;

sub this_ip {
  return $this_ip if defined $this_ip;
  
  my $hostname = hostname();
  my ($name,$aliases,$addrtype,$length,@addrs)  = gethostbyname($hostname);
  foreach (@addrs)
    {
      my @ip = unpack ('C4', $_);
      if ($ip[0] != 127)
	{
	  $this_ip = join '.', @ip;
	  last;
	  }
    }

  return $this_ip;
}

sub get_pkt_proto
  {
    my $pkt = shift;
    my $has_eth = shift;

    return unpack 'c', substr($pkt, (14 * $has_eth) + 9, 1);
  }


sub inet_cksum
  {
    my $msg = shift;

    my ($len_msg,		# Length of the message
        $num_short,		# The number of short words in the message
        $short,			# One short word
        $chk			# The checksum
       );

    $len_msg = length($msg);
    $num_short = int($len_msg / 2);
    $chk = 0;
    foreach $short (unpack("n$num_short", $msg)) {
      $chk += $short;
    }				# Add the odd byte in
    $chk += (unpack("C", substr($msg, $len_msg - 1, 1)) << 8) if $len_msg % 2;
    $chk = ($chk >> 16) + ($chk & 0xffff); # Fold high into low
    return(~(($chk >> 16) + $chk) & 0xffff); # Again and complement
  }


my $gateway_machine = undef; # default if none of the below matches
my @route_entries; # = ([$addrblock, $gateway], ...);

# returns gateway machine for nonlocal machines, the target for local nets
sub get_gateway_machine
  {
    my $target = shift;

    die "if not linux, should not have to call get_gateway_mac_addr"
      if ($^O !~ /linux/);

    if (! defined $gateway_machine)
      {
	my $iface = get_pcap_device();
	my $route_info = `/sbin/route -n | /bin/grep $iface`;

	my @lines = split "\n", $route_info;
	my $header = shift @lines;

	foreach (@lines) {
	  my ($dest, $gateway, $mask) = split /\s+/, $_;

	  # Get MAC addr
	  $gateway = `arp -a $gateway`;
	  ($gateway) = ($gateway =~ /at\s+([0-9A-F:]+)/);

	  if ($dest eq '0.0.0.0' || $dest eq 'default') {
	    $gateway_machine = $gateway;
	  } else {
	    my $block = new2 Net::Netmask($dest, $mask);
	    ( warn ("Problem: Invalid netmask $dest/$mask"), next ) if !$block;
	    push @route_entries, [$block, $gateway];
	  }
	}

	die "Could not find a gateway machine!\n" if ! defined $gateway_machine;
      }

    foreach my $route (@route_entries)
      {
	return $route->[1] if $route->[0]->match($target);
      }

    return $gateway_machine;
  }


1;
