#!/usr/bin/perl

# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

# Active mapping tool for host characterization

use strict;

use Net::RawIP;
use Net::Pcap;

use Data::Dumper;
use Socket;
use ActiveMap::Utils;
use ActiveMap::IP;
use ActiveMap::TCP;
use ActiveMap::Firewall;
use ActiveMap::Network;

sub usage {
  print
'perl activemap.pl [-noip] [-notcp] [-h N] [-t N] lowport highport hosts...

lowport and highport mark the range of TCP ports that have been
firewalled off for use as source ports in mapping.

If -noip is specified, the IP mapping tests are not run.
If -notcp is specified, the TCP mapping tests are not run.
If -h is specified, the next argument says the maximum number
of hosts that will be mapped simultaneously (default = 10)
If -t is specified, the next argument says the maximum number
of tests that will be run simultaneously, counting all hosts (default = 10)

hosts is a list of hostnames and/or IP addresses to map

Output goes to stdout.
';
  exit();
}


#
# arguments: lowport highport hosts
#

my $notcp = 0;
my $noip = 0;

my $max_hosts_at_once = 10;
my $max_tests_at_once = 10;

while ($ARGV[0] =~ /^-/)
  {
    my $arg = shift @ARGV;
    if ($arg eq '-notcp')
      {
	$notcp = 1;
	print STDERR "# Not running TCP tests.\n";
      }
    elsif($arg eq '-noip')
      {
	$noip = 1;
	print STDERR "# Not running IP tests.\n";
      }
    elsif($arg eq '-h')
      {
	$max_hosts_at_once = ((shift @ARGV)+0) || 1;
	print STDERR "# Mapping at most $max_hosts_at_once hosts at once.\n";
      }
    elsif($arg eq '-t')
      {
	$max_tests_at_once = ((shift @ARGV)+0) || 1;
	print STDERR "# Running at most $max_tests_at_once tests at once.\n";
      }
  }

# allocate a port range for tcp tests *** get from cmd line
my $lowport = shift @ARGV or usage();
my $highport = shift @ARGV or usage();
my $tcp_filter = ActiveMap::Firewall::set_port_range($lowport, $highport);

my %addrcache; # name => ip addr
my @hosts = @ARGV;
my @targets;
print STDERR "# Testing for existence of hosts at " . time() . "\n";
foreach (@hosts) {
  my ($name,$aliases,$addrtype,$length,@addrs);
  if ($_ =~ /(\d{1,3}\.){3}\d{1,3}/) { 
    my $iaddr = inet_aton($_); # or whatever address
    ($name,$aliases,$addrtype,$length,@addrs) = gethostbyaddr($iaddr, AF_INET);
  } else {
    ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname($_);
  }

  if (!defined $name)
    {
      print STDERR "# Unknown host $_, skipping.\n";
      next;
    }
  my $ip = inet_ntoa($addrs[0]);
  $addrcache{$ip} = $name;

  push @targets, $ip;
}

print STDERR "# Starting tests at " . time(). "\n";

my $results;
$results = ActiveMap::Network::do_network(@targets);

print STDERR "# Finished network tests at " . time() . ".\n";

$results = ActiveMap::TCP::do_tcp($max_hosts_at_once, $max_tests_at_once, 
				  $results, @targets) unless $notcp;

print STDERR "# Finished TCP tests at " . time() . ".\n" unless $notcp;

$results = ActiveMap::IP::do_ip($max_hosts_at_once, $max_tests_at_once, 
				$results, @targets) unless $noip;

print STDERR "# Finished IP tests at " . time() . ".\n" unless $noip;

foreach (keys %$results)
  {
    $results->{$_}{hostname} = $addrcache{$_} if defined $addrcache{$_};
  }

print Dumper($results);
print STDERR "# Finished printing results at " . time(). "\n";


