#!/usr/bin/perl

# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

use strict;
use Data::Dumper;

# If -i given, don't treat "No results" as inconsistencies
my $ignore_unknown = 0;
my $keep_consistent_only = 0;

my $USAGE = 
'consistent.pl [-i] [-v] file1 file2...

consistent.pl prints out a list of inconsistent host profiles from the
Active Mapping output files specified.

If the -i option is given, then a "No Result" will not count as an
inconsistency.

If the -v option is given, then the output is inverted: only the
consistent profiles will be kept and output in the standard mapper
output format.
';

if (! scalar(@ARGV))
  {
    print $USAGE;
    exit 0;
  }

while ($ARGV[0] =~ /^-/)
  {
    my $arg = shift @ARGV;
    last if ($arg eq '--');

    if ($arg eq '-i') {
      $ignore_unknown = 1;
      shift @ARGV;
    }
    elsif ($arg eq '-v') {
      $keep_consistent_only = 1;
    }
  }

*OUTPUT = *STDOUT;

my @problem_hosts;
my @problem_hostnames;

my %good_hosts;

foreach my $infile (@ARGV) {

  die "Must supply input file name\n" if ! defined $infile;

  $/ = undef;

  open RESULT, $infile or die "could not open input file";

  my $contents = <RESULT>;
  $contents =~ s/^[^\$]*//;	# quick and dirty, get rid of stuff before $VAR1
  my $VAR1;
  my $result = eval ($contents);

  my @tests = ( 'overlapping_frag', 'rst_in_seq',
		'rst_in_window',
		'rst_outside_window',
		'overlapping_seg' );

  # right only supports 1 key per test
 HOST: while ((my $ip, my $profile) = each %$result) {
  TEST: foreach my $test (@tests)
      {
	next if ref ( $profile->{$test} ) ne 'ARRAY';
	my $result = undef;
	foreach my $testresult (@{$profile->{$test}}) {
	  next TEST if (ref($testresult) ne 'HASH');
	  foreach my $key (keys %$testresult) {
	    next if $ignore_unknown && ($testresult->{$key} =~ /no result/i ||
					$testresult->{$key} =~ /did not complete/i);
	    if ( ! defined $result ) {
	      $result = $testresult->{$key};
	      next;
	    } elsif ( $result ne $testresult->{$key} ) # bad: inconsistency
	      {
		print Dumper({ $ip => $profile }) if ! $keep_consistent_only;
		push @problem_hosts, $ip;
		push @problem_hostnames, $profile->{hostname};
		next HOST;
	      }
	  }
	}
      }

   $good_hosts{$ip} = $profile if $keep_consistent_only;
  }

  close RESULT;

}

if (! $keep_consistent_only )
  {
    print "Problematic hosts: \n", 
      join "\n", (map { ($_) . "\t" . (shift @problem_hostnames) } @problem_hosts);
  }
else
  {
    use Data::Dumper;
    print "# Inconsistent hosts filtered out: ".scalar(@problem_hosts)."\n";
    print "# Consistent hosts kept: ".scalar(keys %good_hosts)."\n";
    print Dumper(\%good_hosts);
  }
