#!/usr/bin/perl

# Copyright (c) 2002
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# This code was written by Umesh Shankar <ushankar@cs.berkeley.edu>
# Please contact me with questions, comments, bug fixes, etc.
#

use strict;

sub usage
  {
    print
'Usage:
parse.pl [-c] mapper.out [active.in]

parse.pl takes as input the output of activemap.pl (the first
argument) and outputs (to the optional second argument) a tabular form
of the data where a majority vote or default resolves any
inconsistency across trials.

If -c is specified, then the output is in a format suitable for use as
an active mapping input file in the Bro NIDS.
';
    exit();
  }

# If -c given, output in format for use in Bro; else tabular
my $cmode = 0;

sub valid_filter
  {
    my $res = shift;
    return '' if (($res =~ /unknown/i or $res =~ /no result/i) &&
		  $cmode);
    return $res;
  }

if ($ARGV[0] eq '-c') {
  $cmode = 1;
  shift @ARGV;
}

my ($infile, $outfile) = @ARGV;

usage() if ! defined $infile;

$/ = undef;

# most = take most frequent answer always
# else take the supplied answer if it appears anywhere
my %interp = ( overlapping_frag => { overlap_policy => 'most' },
	       rst_in_seq => { accepts_rst_in_sequence => 'no' },
               rst_in_window => { accepts_rst_in_window => 'no' },
	       rst_outside_window => { accepts_rst_outside_window => 'no' },
               overlapping_seg => { tcp_segment_overlap => 'most'  },
	     );

open RESULT, $infile or die "could not open input file";
if (defined $outfile) {
  open OUTPUT, ">$outfile" or die "could not open output file\n";
} else {
  *OUTPUT = *STDOUT;
}

my $contents = <RESULT>;
$contents =~ s/^[^\$]*//;	# quick and dirty, get rid of stuff before $VAR1
my $VAR1;
my $result = eval ($contents);
my $global_output;
my $valid_count = 0;
my @tests = keys %interp;

my $join_char = $cmode ? "\n" : "\t";
print OUTPUT (scalar(@tests)+2) if $cmode; # +2 for hops, mtu
print OUTPUT "\n";
print OUTPUT ((join $join_char, ("ip", "hops", "mtu", map { keys %{$interp{$_}} } @tests)), "\n");

# right only supports 1 key per test
while ((my $ip, my $profile) = each %$result) {
  my %winners;			# test => winning answer
TEST: foreach my $test (@tests)
    {
      my %freq;
      foreach my $testresult (@{$profile->{$test}}) {
	if (ref($testresult) ne 'HASH') {
	  $winners{$test} = 'No result';
	  next TEST;
	}
	foreach my $key (keys %$testresult) {
	  if ($interp{$test}{$key} eq $testresult->{$key}) {
	    $winners{$test} = $testresult->{$key};
	    next TEST;
	  }
	  next if ($testresult->{$key} =~ /no result/i);
	  $freq{$testresult->{$key}}++;
	}
      }
      my ($most, $value) = (0);
      foreach (keys %freq) {
	if ($freq{$_} > $most) {
	  $most = $freq{$_};
	  $most -= 0.01 if $value =~ /unknown/i; # unknown policy = lower precedence
	  $value = $_;
	}
      }

      $winners{$test} = ($most == 0) ? 'No result' : $value;
      $winners{$test} = valid_filter($winners{$test});
      $winners{$test} =~ s/\n\t//g; # newlines,tabs used as output delimiters
    }

  my $output = valid_filter($profile->{hops}) .  $join_char . 
    valid_filter($profile->{mtu}) .  $join_char;
  $output .= (join $join_char, (map {$winners{$_}} @tests));

  my $outtest = $output;
  $outtest =~ s/$join_char//g;
  if ($outtest ne '')
    {
      $global_output .= "$ip" . $join_char;
      $global_output .= $output;
      $global_output .= "\n";
      $valid_count++;
    }
}

print OUTPUT $valid_count . "\n" if $cmode;
print OUTPUT $global_output;

close OUTPUT;
close RESULT;

